module precision_module
   integer, parameter :: sp_kind     = kind(1.)                  ! single precision
   integer, parameter :: dp_kind     = kind(1.d0)                ! double precision       
   integer, parameter :: my_kind     = sp_kind
   integer, parameter :: kind_higher = selected_real_kind(2*precision(1._my_kind))           ! size of the matrix
contains

   function sum_simple(a,n)
       integer :: n
       real(my_kind) :: a(n)
       real(my_kind) :: sum_simple
       sum_simple = 0._my_kind
       do i=1,n
          sum_simple = sum_simple + a(i)
       enddo
   end function sum_simple
   
   function sum_kahan(a,n)
       integer :: n
       real(my_kind) :: a(n)
! implement Kahan
       real(my_kind) :: s,t,y,z
       s=a(1)
       t=0._my_kind
       do i=1,n
          y = a(i) - t
          z = s + y
          t = (z-s) - y
          s = z
       enddo
       sum_kahan = s
   end function sum_kahan
   
   function sum_higher(a,n)
       integer :: n
       real(my_kind) :: a(n)
       real(my_kind) :: sum_higher
       real(kind_higher) :: sum_higher_temp
       sum_higher_temp = 0._kind_higher
       do i=1,n
          sum_higher_temp = sum_higher_temp + real(a(i),kind_higher)
       enddo
       sum_higher = sum_higher_temp
   end function sum_higher
end module precision_module

program summation

    use precision_module
       
    implicit none

    integer, parameter :: n=10000000! size of the vector
    real(my_kind) :: a(n)
    real(my_kind) :: summ

    call init_values(a,n)

    summ = sum_simple(a,n)
    write(*,'("Summation simple: ",F22.14)') summ
    summ = sum_kahan(a,n)
    write(*,'("Summation Kahan:  ",F22.14)') summ
    summ = sum_higher(a,n)
    write(*,'("Summation Higher: ",F22.14)') summ

end program summation

subroutine init_values(a,n)
    use precision_module
    integer :: n
    real(my_kind) :: a(n)
    real(my_kind) :: temp
    do i=1,n
       call random_number(temp)
       if(mod(i,4) == 0) a(i) =  -100._my_kind * temp
       if(mod(i,4) == 1) a(i) =  0.001_my_kind * temp
       if(mod(i,4) == 2) a(i) =   100._my_kind * temp
       if(mod(i,4) == 3) a(i) = -0.001_my_kind * temp
    enddo
end subroutine init_values
